@@ -3,7 +3,7 @@
 #
 #     This program is under GPL [http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt]
 #
-#     (c) 2007,2008,2009,2010 - Tom Eastep (teastep@shorewall.net)
+#     (c) 2007,2008,2009,2010.2011 - Tom Eastep (teastep@shorewall.net)
 #
 #       Complete documentation is available at http://shorewall.net
 #
@@ -29,13 +29,18 @@ use Shorewall::Config qw(:DEFAULT :internal);
 use Shorewall::IPAddrs;
 use Shorewall::Zones;
 use Shorewall::Chains qw(:DEFAULT :internal);
+use Shorewall::Proc qw( setup_interface_proc );
 
 use strict;
 
 our @ISA = qw(Exporter);
-our @EXPORT = qw( setup_providers @routemarked_interfaces handle_stickiness handle_optional_interfaces );
+our @EXPORT = qw( process_providers
+		  setup_providers
+		  @routemarked_interfaces
+		  handle_stickiness
+		  handle_optional_interfaces );
 our @EXPORT_OK = qw( initialize lookup_provider );
-our $VERSION = '4.4_16';
+our $VERSION = '4.4_23';
 
 use constant { LOCAL_TABLE   => 255,
 	       MAIN_TABLE    => 254,
@@ -43,23 +48,23 @@ use constant { LOCAL_TABLE   => 255,
 	       UNSPEC_TABLE  => 0
 	       };
 
-our @routemarked_providers;
-our %routemarked_interfaces;
+my  @routemarked_providers;
+my  %routemarked_interfaces;
 our @routemarked_interfaces;
-our %provider_interfaces;
+my  %provider_interfaces;
 
-our $balancing;
-our $fallback;
-our $first_default_route;
-our $first_fallback_route;
+my $balancing;
+my $fallback;
+my $first_default_route;
+my $first_fallback_route;
 
-our %providers;
+my %providers;
 
-our @providers;
+my @providers;
 
-our $family;
+my $family;
 
-our $lastmark;
+my $lastmark;
 
 use constant { ROUTEMARKED_SHARED => 1, ROUTEMARKED_UNSHARED => 2 };
 
@@ -85,10 +90,10 @@ sub initialize( $ ) {
     $first_default_route  = 1;
     $first_fallback_route = 1;
 
-    %providers  = ( local   => { number => LOCAL_TABLE   , mark => 0 , optional => 0 } ,
-		    main    => { number => MAIN_TABLE    , mark => 0 , optional => 0 } ,
-		    default => { number => DEFAULT_TABLE , mark => 0 , optional => 0 } ,
-		    unspec  => { number => UNSPEC_TABLE  , mark => 0 , optional => 0 } );
+    %providers  = ( local   => { number => LOCAL_TABLE   , mark => 0 , optional => 0 ,routes => [], rules => [] } ,
+		    main    => { number => MAIN_TABLE    , mark => 0 , optional => 0 ,routes => [], rules => [] } ,
+		    default => { number => DEFAULT_TABLE , mark => 0 , optional => 0 ,routes => [], rules => [] } ,
+		    unspec  => { number => UNSPEC_TABLE  , mark => 0 , optional => 0 ,routes => [], rules => [] } );
     @providers = ();
 }
 
@@ -100,7 +105,7 @@ sub setup_route_marking() {
 
     require_capability( $_ , q(The provider 'track' option) , 's' ) for qw/CONNMARK_MATCH CONNMARK/;
 
-    add_rule $mangle_table->{$_} , "-m connmark ! --mark 0/$mask -j CONNMARK --restore-mark --mask $mask" for qw/PREROUTING OUTPUT/;
+    add_ijump $mangle_table->{$_} , j => 'CONNMARK', targetopts => "--restore-mark --mask $mask", connmark => "! --mark 0/$mask" for qw/PREROUTING OUTPUT/;
 
     my $chainref  = new_chain 'mangle', 'routemark';
     my $chainref1 = new_chain 'mangle', 'setsticky';
@@ -114,22 +119,22 @@ sub setup_route_marking() {
 	my $mark      = $providerref->{mark};
 
 	unless ( $marked_interfaces{$interface} ) {
-	    add_jump $mangle_table->{PREROUTING} , $chainref,  0, "-i $physical -m mark --mark 0/$mask ";
-	    add_jump $mangle_table->{PREROUTING} , $chainref1, 0, "! -i $physical -m mark --mark  $mark/$mask ";
-	    add_jump $mangle_table->{OUTPUT}     , $chainref2, 0, "-m mark --mark  $mark/$mask ";
+	    add_ijump $mangle_table->{PREROUTING} , j => $chainref,  i => $physical,     mark => "--mark 0/$mask";
+	    add_ijump $mangle_table->{PREROUTING} , j => $chainref1, i => "! $physical", mark => "--mark  $mark/$mask";
+	    add_ijump $mangle_table->{OUTPUT}     , j => $chainref2,                     mark => "--mark  $mark/$mask";
 	    $marked_interfaces{$interface} = 1;
 	}
 
 	if ( $providerref->{shared} ) {
 	    add_commands( $chainref, qq(if [ -n "$providerref->{mac}" ]; then) ), incr_cmd_level( $chainref ) if $providerref->{optional};
-	    add_rule $chainref, match_source_dev( $interface ) . "-m mac --mac-source $providerref->{mac} -j MARK --set-mark $providerref->{mark}";
+	    add_ijump $chainref, j => 'MARK', targetopts => "--set-mark $providerref->{mark}", imatch_source_dev( $interface ), mac => "--mac-source $providerref->{mac}";
 	    decr_cmd_level( $chainref ), add_commands( $chainref, "fi\n" ) if $providerref->{optional};
 	} else {
-	    add_rule $chainref, match_source_dev( $interface ) . "-j MARK --set-mark $providerref->{mark}";
+	    add_ijump $chainref, j => 'MARK', targetopts => "--set-mark $providerref->{mark}", imatch_source_dev( $interface );
 	}
     }
 
-    add_rule $chainref, "-m mark ! --mark 0/$mask -j CONNMARK --save-mark --mask $mask";
+    add_ijump $chainref, j => 'CONNMARK', targetopts => "--save-mark --mask $mask", mark => "! --mark 0/$mask";
 }
 
 sub copy_table( $$$ ) {
@@ -139,14 +144,16 @@ sub copy_table( $$$ ) {
     #
     my $filter = $family == F_IPV6 ? q(sed 's/ via :: / /' | ) : '';
 
+    emit '';
+
     if ( $realm ) {
-	emit  ( "\$IP -$family route show table $duplicate | sed -r 's/ realm [[:alnum:]_]+//' | while read net route; do" )
+	emit  ( "\$IP -$family -o route show table $duplicate | sed -r 's/ realm [[:alnum:]_]+//' | while read net route; do" )
     } else {
-	emit  ( "\$IP -$family route show table $duplicate | ${filter}while read net route; do" )
+	emit  ( "\$IP -$family -o route show table $duplicate | ${filter}while read net route; do" )
     }
 
     emit ( '    case $net in',
-	   '        default|nexthop)',
+	   '        default)',
 	   '            ;;',
 	   '        *)',
 	   "            run_ip route add table $number \$net \$route $realm",
@@ -170,15 +177,17 @@ sub copy_and_edit_table( $$$$ ) {
     # Shell and iptables use a different wildcard character
     #
     $copy =~ s/\+/*/;
+    
+    emit '';
 
     if ( $realm ) {
-	emit  ( "\$IP -$family route show table $duplicate | sed -r 's/ realm [[:alnum:]]+//' | while read net route; do" )
+	emit  ( "\$IP -$family -o route show table $duplicate | sed -r 's/ realm [[:alnum:]]+//' | while read net route; do" )
     } else {
-	emit  ( "\$IP -$family route show table $duplicate | ${filter}while read net route; do" )
+	emit  ( "\$IP -$family -o route show table $duplicate | ${filter}while read net route; do" )
     }
 
     emit (  '    case $net in',
-	    '        default|nexthop)',
+	    '        default)',
 	    '            ;;',
 	    '        *)',
 	    '            case $(find_device $route) in',
@@ -242,21 +251,28 @@ sub balance_fallback_route( $$$$ ) {
 sub start_provider( $$$ ) {
     my ($table, $number, $test ) = @_;
 
+    emit "\n#\n# Add Provider $table ($number)\n#";
+
+    emit "start_provider_$table() {";
+    push_indent;
     emit $test;
     push_indent;
 
-    emit "#\n# Add Provider $table ($number)\n#";
-
     emit "qt ip -$family route flush table $number";
-    emit "echo \"qt \$IP -$family route flush table $number\" >> \${VARDIR}/undo_routing";
+    emit "echo \"qt \$IP -$family route flush table $number\" > \${VARDIR}/undo_${table}_routing";
 }
 
-sub add_a_provider( ) {
+#
+# Process a record in the providers file
+#
+sub process_a_provider() {
 
     my ($table, $number, $mark, $duplicate, $interface, $gateway,  $options, $copy ) = split_line 6, 8, 'providers file';
 
     fatal_error "Duplicate provider ($table)" if $providers{$table};
 
+    fatal_error "Invalid Provider Name ($table)" unless $table =~ /^[\w]+$/;
+
     my $num = numeric_value $number;
 
     fatal_error "Invalid Provider number ($number)" unless defined $num;
@@ -281,7 +297,6 @@ sub add_a_provider( ) {
     fatal_error "A bridge port ($interface) may not be configured as a provider interface" if port_to_bridge $interface;
 
     my $physical    = get_physical $interface;
-    my $base        = uc chain_base $physical;
     my $gatewaycase = '';
 
     if ( $gateway eq 'detect' ) {
@@ -303,6 +318,7 @@ sub add_a_provider( ) {
     unless ( $options eq '-' ) {
 	for my $option ( split_list $options, 'option' ) {
 	    if ( $option eq 'track' ) {
+		require_capability( 'MANGLE_ENABLED' , q(The 'track' option) , 's' );
 		$track = 1;
 	    } elsif ( $option eq 'notrack' ) {
 		$track = 0;
@@ -349,6 +365,15 @@ sub add_a_provider( ) {
 	}
     }
 
+    fatal_error q(The 'balance' and 'fallback' options are mutually exclusive) if $balance && $default;
+
+    if ( $local ) {
+	fatal_error "GATEWAY not valid with 'local' provider" unless $gatewaycase eq 'none';
+	fatal_error "'track' not valid with 'local'"          if $track;
+	fatal_error "DUPLICATE not valid with 'local'"        if $duplicate ne '-';
+	fatal_error "MARK required with 'local'"              unless $mark;
+    }
+
     my $val = 0;
     my $pref;
 
@@ -356,6 +381,8 @@ sub add_a_provider( ) {
 
     if ( $mark ne '-' ) {
 
+	require_capability( 'MANGLE_ENABLED' , 'Provider marks' , '' );
+
 	$val = numeric_value $mark;
 
 	fatal_error "Invalid Mark Value ($mark)" unless defined $val && $val;
@@ -383,8 +410,18 @@ sub add_a_provider( ) {
 
     $balance = $default_balance unless $balance;
 
+    fatal_error "Interface $interface is already associated with non-shared provider $provider_interfaces{$interface}" if $provider_interfaces{$table};
+
+    if ( $duplicate ne '-' ) {
+	fatal_error "The DUPLICATE column must be empty when USE_DEFAULT_RT=Yes" if $config{USE_DEFAULT_RT};
+    } elsif ( $copy ne '-' ) {
+	fatal_error "The COPY column must be empty when USE_DEFAULT_RT=Yes" if $config{USE_DEFAULT_RT};
+	fatal_error 'A non-empty COPY column requires that a routing table be specified in the DUPLICATE column';
+    }
+
     $providers{$table} = { provider    => $table,
 			   number      => $number ,
+			   rawmark     => $mark ,
 			   mark        => $val ? in_hex($val) : $val ,
 			   interface   => $interface ,
 			   physical    => $physical ,
@@ -392,7 +429,19 @@ sub add_a_provider( ) {
 			   gateway     => $gateway ,
 			   gatewaycase => $gatewaycase ,
 			   shared      => $shared ,
-			   default     => $default };
+			   default     => $default ,
+			   copy        => $copy ,
+			   balance     => $balance ,
+			   pref        => $pref ,
+			   mtu         => $mtu ,
+			   track       => $track ,
+			   loose       => $loose ,
+			   duplicate   => $duplicate ,
+			   address     => $address ,
+			   local       => $local ,
+			   rules       => [] ,
+			   routes      => [] ,
+			 };
 
     if ( $track ) {
 	fatal_error "The 'track' option requires a numeric value in the MARK column" if $mark eq '-';
@@ -408,9 +457,39 @@ sub add_a_provider( ) {
 	push @routemarked_providers, $providers{$table};
     }
 
-    my $realm = '';
+    push @providers, $table;
 
-    fatal_error "Interface $interface is already associated with non-shared provider $provider_interfaces{$interface}" if $provider_interfaces{$table};
+}
+
+#
+# Generate the start_provider_...() function for the passed provider
+#
+sub add_a_provider( $$ ) {
+  
+    my ( $providerref, $tcdevices ) = @_;
+    
+    my $table       = $providerref->{provider};
+    my $number      = $providerref->{number};
+    my $mark        = $providerref->{rawmark};
+    my $interface   = $providerref->{interface};
+    my $physical    = $providerref->{physical};
+    my $optional    = $providerref->{optional};
+    my $gateway     = $providerref->{gateway};
+    my $gatewaycase = $providerref->{gatewaycase};
+    my $shared      = $providerref->{shared};
+    my $default     = $providerref->{default};
+    my $copy        = $providerref->{copy};
+    my $balance     = $providerref->{balance};
+    my $pref        = $providerref->{pref};
+    my $mtu         = $providerref->{mtu};
+    my $track       = $providerref->{track};
+    my $loose       = $providerref->{loose};
+    my $duplicate   = $providerref->{duplicate};
+    my $address     = $providerref->{address};
+    my $local       = $providerref->{local};
+    my $dev         = chain_base $physical;
+    my $base        = uc $dev;
+    my $realm = '';
 
     if ( $shared ) {
 	my $variable = $providers{$table}{mac} = get_interface_mac( $gateway, $interface , $table );
@@ -424,7 +503,6 @@ sub add_a_provider( ) {
 	} else {
 	    start_provider( $table, $number, "if interface_is_usable $physical; then" );
 	}
-
 	$provider_interfaces{$interface} = $table;
 
 	if ( $gatewaycase eq 'none' ) {
@@ -435,6 +513,11 @@ sub add_a_provider( ) {
 	    }
 	}
     }
+    
+    #
+    # /proc for this interface
+    #
+    setup_interface_proc( $interface );
 
     if ( $mark ne '-' ) {
 	my $mask = have_capability 'FWMARK_RT_MASK' ? '/' . in_hex $globals{PROVIDER_MASK} : '';
@@ -442,12 +525,11 @@ sub add_a_provider( ) {
 	emit ( "qt \$IP -$family rule del fwmark ${mark}${mask}" ) if $config{DELETE_THEN_ADD};
 
 	emit ( "run_ip rule add fwmark ${mark}${mask} pref $pref table $number",
-	       "echo \"qt \$IP -$family rule del fwmark ${mark}${mask}\" >> \${VARDIR}/undo_routing"
+	       "echo \"qt \$IP -$family rule del fwmark ${mark}${mask}\" >> \${VARDIR}/undo_${table}_routing"
 	     );
     }
 
     if ( $duplicate ne '-' ) {
-	fatal_error "The DUPLICATE column must be empty when USE_DEFAULT_RT=Yes" if $config{USE_DEFAULT_RT};
 	if ( $copy eq '-' ) {
 	    copy_table ( $duplicate, $number, $realm );
 	} else {
@@ -459,109 +541,187 @@ sub add_a_provider( ) {
 
 	    copy_and_edit_table( $duplicate, $number ,$copy , $realm);
 	}
-    } elsif ( $copy ne '-' ) {
-	fatal_error "The COPY column must be empty when USE_DEFAULT_RT=Yes" if $config{USE_DEFAULT_RT};
-	fatal_error 'A non-empty COPY column requires that a routing table be specified in the DUPLICATE column';
     }
 
     if ( $gateway ) {
 	$address = get_interface_address $interface unless $address;
-	emit "run_ip route replace $gateway src $address dev $physical ${mtu}table $number $realm";
+	if ( $family == F_IPV4 ) {
+	    emit "run_ip route replace $gateway src $address dev $physical ${mtu}";
+	    emit "run_ip route replace $gateway src $address dev $physical ${mtu}table $number $realm";
+	} else {
+	    emit "qt \$IP -6 route del $gateway src $address dev $physical ${mtu}";
+	    emit "run_ip route add $gateway src $address dev $physical ${mtu}";
+	    emit "qt \$IP -6 route del $gateway src $address dev $physical ${mtu}table $number $realm";
+	    emit "run_ip route add $gateway src $address dev $physical ${mtu}table $number $realm";
+	}
+   	
 	emit "run_ip route add default via $gateway src $address dev $physical ${mtu}table $number $realm";
     }
 
-    balance_default_route $balance , $gateway, $physical, $realm if $balance;
+    balance_default_route( $balance , $gateway, $physical, $realm ) if $balance;
 
     if ( $default > 0 ) {
-	balance_fallback_route $default , $gateway, $physical, $realm;
+	balance_fallback_route( $default , $gateway, $physical, $realm );
     } elsif ( $default ) {
 	emit '';
 	if ( $gateway ) {
-	    emit qq(run_ip route replace default via $gateway src $address dev $physical table ) . DEFAULT_TABLE . qq( metric $number);
-	    emit qq(echo "qt \$IP -$family route del default via $gateway table ) . DEFAULT_TABLE . qq(" >> \${VARDIR}/undo_routing);
+	    if ( $family == F_IPV4 ) {
+		emit qq(run_ip route replace default via $gateway src $address dev $physical table ) . DEFAULT_TABLE . qq( metric $number);
+	    } else {
+		emit qq(qt \$IP -6 route del default via $gateway src $address dev $physical table ) . DEFAULT_TABLE . qq( metric $number);
+		emit qq(run_ip route add default via $gateway src $address dev $physical table ) . DEFAULT_TABLE . qq( metric $number);
+	    }
+	    emit qq(echo "qt \$IP -$family route del default via $gateway table ) . DEFAULT_TABLE . qq(" >> \${VARDIR}/undo_${table}_routing);
 	} else {
 	    emit qq(run_ip route add default table ) . DEFAULT_TABLE . qq( dev $physical metric $number);
-	    emit qq(echo "qt \$IP -$family route del default dev $physical table ) . DEFAULT_TABLE . qq(" >> \${VARDIR}/undo_routing);
+	    emit qq(echo "qt \$IP -$family route del default dev $physical table ) . DEFAULT_TABLE . qq(" >> \${VARDIR}/undo_${table}_routing);
 	}
     }
 
-    if ( $local ) {
-	fatal_error "GATEWAY not valid with 'local' provider" unless $gatewaycase eq 'none';
-	fatal_error "'track' not valid with 'local'"          if $track;
-	fatal_error "DUPLICATE not valid with 'local'"        if $duplicate ne '-';
-	fatal_error "MARK required with 'local'"              unless $mark;
-    } elsif ( $loose ) {
-	if ( $config{DELETE_THEN_ADD} ) {
-	    emit ( "\nfind_interface_addresses $physical | while read address; do",
-		   "    qt \$IP -$family rule del from \$address",
-		   'done'
-		 );
+    unless ( $local ) {
+	if ( $loose ) {
+	    if ( $config{DELETE_THEN_ADD} ) {
+		emit ( "\nfind_interface_addresses $physical | while read address; do",
+		       "    qt \$IP -$family rule del from \$address",
+		       'done'
+		     );
+	    }
+	} elsif ( $shared ) {
+	    emit  "qt \$IP -$family rule del from $address" if $config{DELETE_THEN_ADD};
+	    emit( "run_ip rule add from $address pref 20000 table $number" ,
+		  "echo \"qt \$IP -$family rule del from $address\" >> \${VARDIR}/undo_${table}_routing" );
+	} else {
+	    my $rulebase = 20000 + ( 256 * ( $number - 1 ) );
+
+	    emit "\nrulenum=$rulebase\n";
+
+	    emit  ( "find_interface_addresses $physical | while read address; do" );
+	    emit  ( "    qt \$IP -$family rule del from \$address" ) if $config{DELETE_THEN_ADD};
+	    emit  ( "    run_ip rule add from \$address pref \$rulenum table $number",
+		    "    echo \"qt \$IP -$family rule del from \$address\" >> \${VARDIR}/undo_${table}_routing",
+		    '    rulenum=$(($rulenum + 1))',
+		    'done'
+		  );
 	}
-    } elsif ( $shared ) {
-	emit  "qt \$IP -$family rule del from $address" if $config{DELETE_THEN_ADD};
-	emit( "run_ip rule add from $address pref 20000 table $number" ,
-	      "echo \"qt \$IP -$family rule del from $address\" >> \${VARDIR}/undo_routing" );
-    } else {
-	my $rulebase = 20000 + ( 256 * ( $number - 1 ) );
-
-	emit "\nrulenum=0\n";
-
-	emit  ( "find_interface_addresses $physical | while read address; do" );
-	emit  (	"    qt \$IP -$family rule del from \$address" ) if $config{DELETE_THEN_ADD};
-	emit  (	"    run_ip rule add from \$address pref \$(( $rulebase + \$rulenum )) table $number",
-		"    echo \"qt \$IP -$family rule del from \$address\" >> \${VARDIR}/undo_routing",
-		'    rulenum=$(($rulenum + 1))',
-		'done'
-	      );
     }
 
-    emit qq(\nprogress_message "   Provider $table ($number) Added"\n);
+    if ( @{$providerref->{rules}} ) {
+	emit '';
+	emit $_ for @{$providers{$table}->{rules}};
+    }
+    
+    if ( @{$providerref->{routes}} ) {
+	emit '';
+	emit $_ for @{$providers{$table}->{routes}};
+    }
+
+    emit( '',
+	  'if [ $COMMAND = enable ]; then'
+	);
+
+    push_indent;
+
+    my ( $tbl, $weight ); 
+    
+    if ( $balance || $default ) {
+	$tbl    = $default || $config{USE_DEFAULT_RT} ? DEFAULT_TABLE : MAIN_TABLE;
+	$weight = $balance ? $balance : $default;
+
+	if ( $gateway ) {
+	    emit qq(add_gateway "nexthop via $gateway dev $physical weight $weight $realm" ) . $tbl;
+	} else {
+	    emit qq(add_gateway "nexthop dev $physical weight $weight $realm" ) . $tbl;
+	}
+
+    }
+
+    emit( "setup_${dev}_tc" ) if $tcdevices->{$interface};
+
+    emit ( qq(progress_message2 "   Provider $table ($number) Started") );
 
     pop_indent;
+	  
+    emit( 'else',
+	  qq(    echo $weight > \${VARDIR}/${physical}_weight),
+	  qq(    progress_message "   Provider $table ($number) Started"),
+	  "fi\n"
+	);
+
+    pop_indent;
+
     emit 'else';
 
+    push_indent;
+
     if ( $optional ) {
 	if ( $shared ) {
-	    emit ( "    error_message \"WARNING: Gateway $gateway is not reachable -- Provider $table ($number) not Added\"" );	    
+	    emit ( "error_message \"WARNING: Gateway $gateway is not reachable -- Provider $table ($number) not Started\"" );	    
 	} else {
-	    emit ( "    error_message \"WARNING: Interface $physical is not usable -- Provider $table ($number) not Added\"" );
+	    emit ( "error_message \"WARNING: Interface $physical is not usable -- Provider $table ($number) not Started\"" );
 	}
     } else {
 	if ( $shared ) {
-	    emit( "    fatal_error \"Gateway $gateway is not reachable -- Provider $table ($number) Cannot be Added\"" );
+	    emit( "fatal_error \"Gateway $gateway is not reachable -- Provider $table ($number) Cannot be Started\"" );
 	} else {
-	    emit( "    fatal_error \"Interface $physical is not usable -- Provider $table ($number) Cannot be Added\"" );
+	    emit( "fatal_error \"Interface $physical is not usable -- Provider $table ($number) Cannot be Started\"" );
 	}
     }
 
-    emit "fi\n";
+    pop_indent;
 
-    push @providers, $table;
+    emit 'fi';
+
+    pop_indent;
+
+    emit '}'; # End of start_provider_$table();
+
+    if ( $optional ) {
+	emit( '',
+	      '#',
+	      "# Stop provider $table",
+	      '#',
+	      "stop_provider_$table() {" );
+
+	push_indent;
+
+	my $undo = "\${VARDIR}/undo_${table}_routing";
+
+	emit( "if [ -f $undo ]; then",
+	      "    . $undo",
+	      "    > $undo" );
+
+	if ( $balance || $default ) {
+	    $tbl    = $fallback || ( $config{USE_DEFAULT_RT} ? DEFAULT_TABLE : MAIN_TABLE );
+	    $weight = $balance ? $balance : $default;
+
+	    my $via = 'via';
+
+	    $via .= " $gateway"       if $gateway;
+	    $via .= " dev $physical";
+	    $via .= " weight $weight";
+	    $via .= " $realm"         if $realm;
+
+	    emit( qq(    delete_gateway "$via" $tbl $physical) );
+	}
+	
+	emit( '', 
+	      "    qt \$TC qdisc del dev $physical root",
+	      "    qt \$TC qdisc del dev $physical ingress\n" ) if $tcdevices->{$interface};
+
+	emit( "    progress_message2 \"Provider $table stopped\"",
+              'else',
+	      "    startup_error \"$undo does not exist\"",
+	      'fi'
+	    );
+
+	pop_indent;
+
+	emit '}';
+    }
 
     progress_message "   Provider \"$currentline\" $done";
 }
 
-#
-# Begin an 'if' statement testing whether the passed interface is available
-#
-sub start_new_if( $ ) {
-    our $current_if = shift;
-
-    emit ( '', qq(if [ -n "\$SW_${current_if}_IS_USABLE" ]; then) );
-    push_indent;
-}
-
-#
-# Complete any current 'if' statement in the output script
-#
-sub finish_current_if() {
-    if ( our $current_if ) {
-	pop_indent;
-	emit ( "fi\n" );
-	$current_if = '';
-    }
-}
-
 sub add_an_rtrule( ) {
     my ( $source, $dest, $provider, $priority ) = split_line 4, 4, 'route_rules file';
 
@@ -585,6 +745,11 @@ sub add_an_rtrule( ) {
 	fatal_error "Unknown provider ($provider)" unless $found;
     }
 
+    my $providerref = $providers{$provider};
+
+    my $number = $providerref->{number};
+
+    fatal_error "You may not add rules for the $provider provider" if $number == LOCAL_TABLE || $number == UNSPEC_TABLE;
     fatal_error "You must specify either the source or destination in a route_rules entry" if $source eq '-' && $dest eq '-';
 
     if ( $dest eq '-' ) {
@@ -625,20 +790,9 @@ sub add_an_rtrule( ) {
 
     $priority = "priority $priority";
 
-    finish_current_if, emit ( "qt \$IP -$family rule del $source $dest $priority" ) if $config{DELETE_THEN_ADD};
-
-    my ( $optional, $number ) = ( $providers{$provider}{optional} , $providers{$provider}{number} );
-
-    if ( $optional ) {
-	my $base = uc chain_base( $providers{$provider}{physical} );
-	finish_current_if if $base ne $current_if;
-	start_new_if( $base ) unless $current_if;
-    } else {
-	finish_current_if;
-    }
-
-    emit ( "run_ip rule add $source $dest $priority table $number",
-	   "echo \"qt \$IP -$family rule del $source $dest $priority\" >> \${VARDIR}/undo_routing" );
+    push @{$providerref->{rules}}, "qt \$IP -$family rule del $source $dest $priority" if $config{DELETE_THEN_ADD};
+    push @{$providerref->{rules}}, "run_ip rule add $source $dest $priority table $number";
+    push @{$providerref->{rules}}, "echo \"qt \$IP -$family rule del $source $dest $priority\" >> \${VARDIR}/undo_${provider}_routing";
 
     progress_message "   Routing rule \"$currentline\" $done";
 }
@@ -657,7 +811,6 @@ sub add_a_route( ) {
 	    for ( keys %providers ) {
 		if ( $providers{$_}{number} == $provider_number ) {
 		    $provider = $_;
-		    fatal_error "You may not add routes to the $provider table" if $provider_number == LOCAL_TABLE || $provider_number == UNSPEC_TABLE;
 		    $found = 1;
 		    last;
 		}
@@ -671,30 +824,25 @@ sub add_a_route( ) {
 
     validate_address ( $gateway, 1 ) if $gateway ne '-';
 
-    my ( $optional, $number ) = ( $providers{$provider}{optional} , $providers{$provider}{number} );
-
+    my $providerref = $providers{$provider};
+    my $number = $providerref->{number};
     my $physical = $device eq '-' ? $providers{$provider}{physical} : physical_name( $device );
-    
-    if ( $providers{$provider}{optional} ) {
-	my $base = uc chain_base( $physical );
-	finish_current_if if $base ne $current_if;
-	start_new_if ( $base ) unless $current_if;
-    } else {
-	finish_current_if;
-    }
+    my $routes = $providerref->{routes};
 
+    fatal_error "You may not add routes to the $provider table" if $number == LOCAL_TABLE || $number == UNSPEC_TABLE;
+    
     if ( $gateway ne '-' ) {
 	if ( $device ne '-' ) {
-	    emit qq(run_ip route add $dest via $gateway dev $physical table $number);
-	    emit qq(echo "qt \$IP -$family route del $dest via $gateway dev $physical table $number" >> \${VARDIR}/undo_routing) if $number >= DEFAULT_TABLE;
+	    push @$routes, qq(run_ip route add $dest via $gateway dev $physical table $number);
+	    emit qq(echo "qt \$IP -$family route del $dest via $gateway dev $physical table $number" >> \${VARDIR}/undo_${provider}_routing) if $number >= DEFAULT_TABLE;
 	} else {
-	    emit qq(run_ip route add $dest via $gateway table $number);
-	    emit qq(echo "\$IP -$family route del $dest via $gateway table $number" >> \${VARDIR}/undo_routing) if $number >= DEFAULT_TABLE; 
+	    push @$routes, qq(run_ip route add $dest via $gateway table $number);
+	    emit qq(echo "\$IP -$family route del $dest via $gateway table $number" >> \${VARDIR}/undo_${provider}_routing) if $number >= DEFAULT_TABLE; 
 	}
     } else {
 	fatal_error "You must specify a device for this route" unless $physical;
-	emit qq(run_ip route add $dest dev $physical table $number);
-	emit qq(echo "\$IP -$family route del $dest dev $physical table $number" >> \${VARDIR}/undo_routing) if $number >= DEFAULT_TABLE;
+	push @$routes, qq(run_ip route add $dest dev $physical table $number);
+	emit qq(echo "\$IP -$family route del $dest dev $physical table $number" >> \${VARDIR}/undo_${provider}_routing) if $number >= DEFAULT_TABLE;
     }
 
     progress_message "   Route \"$currentline\" $done";
@@ -702,17 +850,16 @@ sub add_a_route( ) {
 
 sub setup_null_routing() {
     save_progress_message "Null Routing the RFC 1918 subnets";
+    emit "> \${VARDIR}undo_rfc1918_routing\n";
     for ( rfc1918_networks ) {
 	emit( qq(if ! \$IP -4 route ls | grep -q '^$_.* dev '; then),
 	      qq(    run_ip route replace unreachable $_),
-	      qq(    echo "qt \$IP -4 route del unreachable $_" >> \${VARDIR}/undo_routing),
+	      qq(    echo "qt \$IP -4 route del unreachable $_" >> \${VARDIR}/undo_rfc1918_routing),
 	      qq(fi\n) );
     }
 }
 
 sub start_providers() {
-    require_capability( 'MANGLE_ENABLED' , 'a non-empty providers file' , 's' );
-
     emit  ( '#',
 	    '# Undo any changes made since the last time that we [re]started -- this will not restore the default route',
 	    '#',
@@ -730,17 +877,22 @@ sub start_providers() {
     emit  ( '#',
 	    '# Capture the default route(s) if we don\'t have it (them) already.',
 	    '#',
-	    "[ -f \${VARDIR}/default_route ] || \$IP -$family route list | save_default_route > \${VARDIR}/default_route",
-	    '#',
-	    '# Initialize the file that holds \'undo\' commands',
-	    '#',
-	    '> ${VARDIR}/undo_routing' );
+	    "[ -f \${VARDIR}/default_route ] || \$IP -$family route list | save_default_route > \${VARDIR}/default_route" );
 
     save_progress_message 'Adding Providers...';
 
     emit 'DEFAULT_ROUTE=';
     emit 'FALLBACK_ROUTE=';
     emit '';
+    
+    for my $provider ( qw/main default/ ) {
+	emit '';
+	emit qq(> \${VARDIR}/undo_${provider}_routing );
+	emit '';
+	emit $_ for @{$providers{$provider}{routes}};
+	emit '';
+	emit $_ for @{$providers{$provider}{rules}};
+    }
 }
 
 sub finish_providers() {
@@ -750,14 +902,19 @@ sub finish_providers() {
 	if ( $config{USE_DEFAULT_RT} ) {
 	    emit ( 'run_ip rule add from ' . ALLIP . ' table ' . MAIN_TABLE . ' pref 999',
 		   "\$IP -$family rule del from " . ALLIP . ' table ' . MAIN_TABLE . ' pref 32766',
-		   qq(echo "qt \$IP -$family rule add from ) . ALLIP . ' table ' . MAIN_TABLE . ' pref 32766" >> ${VARDIR}/undo_routing',
-		   qq(echo "qt \$IP -$family rule del from ) . ALLIP . ' table ' . MAIN_TABLE . ' pref 999" >> ${VARDIR}/undo_routing',
+		   qq(echo "qt \$IP -$family rule add from ) . ALLIP . ' table ' . MAIN_TABLE . ' pref 32766" >> ${VARDIR}/undo_main_routing',
+		   qq(echo "qt \$IP -$family rule del from ) . ALLIP . ' table ' . MAIN_TABLE . ' pref 999" >> ${VARDIR}/undo_main_routing',
 		   '' );
 	    $table = DEFAULT_TABLE;
 	}
 
 	emit  ( 'if [ -n "$DEFAULT_ROUTE" ]; then' );
-	emit  ( "    run_ip route replace default scope global table $table \$DEFAULT_ROUTE" );
+	if ( $family == F_IPV4 ) {
+	    emit  ( "    run_ip route replace default scope global table $table \$DEFAULT_ROUTE" );
+	} else {
+	    emit  ( "    qt \$IP -6 route del default scope global table $table \$DEFAULT_ROUTE" );
+	    emit  ( "    run_ip route add default scope global table $table \$DEFAULT_ROUTE" );
+	}
 
 	if ( $config{USE_DEFAULT_RT} ) {
 	    emit  ( "    while qt \$IP -$family route del default table " . MAIN_TABLE . '; do',
@@ -788,11 +945,17 @@ sub finish_providers() {
     }
 
     if ( $fallback ) {
-	emit  ( 'if [ -n "$FALLBACK_ROUTE" ]; then' ,
-		"    run_ip route replace default scope global table " . DEFAULT_TABLE . " \$FALLBACK_ROUTE" ,
-		"    progress_message \"Fallback route '\$(echo \$FALLBACK_ROUTE | sed 's/\$\\s*//')' Added\"",
-		'fi',
-		'' );
+	emit  ( 'if [ -n "$FALLBACK_ROUTE" ]; then' );
+	if ( $family == F_IPV4 ) {
+	    emit( "    run_ip route replace default scope global table " . DEFAULT_TABLE . " \$FALLBACK_ROUTE" );
+	} else {
+	    emit( "    qt \$IP -6 route del default scope global table " . DEFAULT_TABLE . " \$FALLBACK_ROUTE" );
+	    emit( "    run_ip route add default scope global table " . DEFAULT_TABLE . " \$FALLBACK_ROUTE" );
+	}
+
+	emit( "    progress_message \"Fallback route '\$(echo \$FALLBACK_ROUTE | sed 's/\$\\s*//')' Added\"",
+	      'fi',
+	      '' );
     }
 
     unless ( $config{KEEP_RT_TABLES} ) {
@@ -817,58 +980,136 @@ sub finish_providers() {
     }
 }
 
-sub setup_providers() {
-    my $providers = 0;
+sub process_providers( $ ) {
+    my $tcdevices = shift;
+
+    our $providers = 0;
 
     $lastmark = 0;
 
     if ( my $fn = open_file 'providers' ) {
-
-	first_entry sub() {
-	    progress_message2 "$doing $fn...";
-	    emit "\nif [ -z \"\$g_noroutes\" ]; then";
-	    push_indent;
-	    start_providers; };
- 
-	add_a_provider, $providers++ while read_a_line;
+	first_entry "$doing $fn..."; 
+	process_a_provider, $providers++ while read_a_line;
     }
 
     if ( $providers ) {
-	finish_providers;
-
-	my $fn = open_file 'routes';
+	my $fn = open_file 'route_rules';
 
 	if ( $fn ) {
-	    our $current_if = '';
-
 	    first_entry "$doing $fn...";
-
-	    emit '';
-
-	    add_a_route while read_a_line;
-
-	    finish_current_if;
-	}
-
-	$fn = open_file 'route_rules';
-
-	if ( $fn ) {
-	    our $current_if = '';
-
-	    first_entry "$doing $fn...";
-
+	    
 	    emit '';
 
 	    add_an_rtrule while read_a_line;
-
-	    finish_current_if;
 	}
 
+	$fn = open_file 'routes';
+
+	if ( $fn ) {
+	    first_entry "$doing $fn...";
+	    emit '';
+	    add_a_route while read_a_line;
+	}
+    }
+
+    add_a_provider( $providers{$_}, $tcdevices ) for @providers;
+    
+    emit << 'EOF';;
+
+#
+# Enable an optional provider
+#
+enable_provider() {
+    g_interface=$1;
+
+    case $g_interface in
+EOF
+
+    push_indent;
+    push_indent;
+
+    for my $provider (@providers ) {
+	my $providerref = $providers{$provider};
+
+	emit( "$providerref->{physical})",
+	      "    if [ -z \"`\$IP -$family route ls table $providerref->{number}`\" ]; then", 
+	      "        start_provider_$provider",
+	      '    else',
+	      '        startup_error "Interface $g_interface is already enabled"',
+	      '    fi',
+	      '    ;;'
+	    ) if $providerref->{optional};
+    }
+
+    pop_indent;
+    pop_indent;
+
+    emit << 'EOF';;
+        *)
+            startup_error "$g_interface is not an optional provider interface"
+            ;;
+    esac
+}
+
+#
+# Disable an optional provider
+#
+disable_provider() {
+    g_interface=$1;
+
+    case $g_interface in
+EOF
+
+    push_indent;
+    push_indent;
+
+    for my $provider (@providers ) {
+	my $providerref = $providers{$provider};
+
+	emit( "$providerref->{physical})",
+	      "    if [ -n \"`\$IP -$family route ls table $providerref->{number}`\" ]; then", 
+	      "        stop_provider_$provider",
+	      '    else',
+	      '        startup_error "Interface $g_interface is already disabled"',
+	      '    fi',
+	      '    ;;'
+	    ) if $providerref->{optional};
+    }
+
+    pop_indent;
+    pop_indent;
+
+    emit << 'EOF';;
+        *)
+            startup_error "$g_interface is not an optional provider interface"
+            ;;
+    esac
+}
+EOF
+
+}
+
+sub setup_providers() {
+    our $providers;
+
+    if ( $providers ) {
+	emit "\nif [ -z \"\$g_noroutes\" ]; then";
+	
+	push_indent;
+
+	start_providers;
+	
+	emit '';
+
+	emit "start_provider_$_" for @providers;
+
+	emit '';
+
+	finish_providers;
+
 	setup_null_routing if $config{NULL_ROUTE_RFC1918};
 	emit "\nrun_ip route flush cache";
-	#
-	# This completes the if-block begun in the first_entry closure above
-	#
+
 	pop_indent;
 	emit "fi\n";
 
@@ -882,10 +1123,6 @@ sub setup_providers() {
 	emit "restore_default_route $config{USE_DEFAULT_RT}";
 
 	if ( $config{NULL_ROUTE_RFC1918} ) {
-	    emit  ( '#',
-		    '# Initialize the file that holds \'undo\' commands',
-		    '#',
-		    '> ${VARDIR}/undo_routing' );
 	    setup_null_routing;
 	    emit "\nrun_ip route flush cache";
 	}
@@ -1068,7 +1305,7 @@ sub handle_stickiness( $ ) {
 	    my $base      = uc chain_base $interface;
 	    my $mark      = $providerref->{mark};
 
-	    for ( grep /-j sticky/, @{$tcpreref->{rules}} ) {
+	    for ( grep rule_target($_) eq 'sticky', @{$tcpreref->{rules}} ) {
 		my $stickyref = ensure_mangle_chain 'sticky';
 		my ( $rule1, $rule2 );
 		my $list = sprintf "sticky%03d" , $sticky++;
@@ -1076,26 +1313,32 @@ sub handle_stickiness( $ ) {
 		for my $chainref ( $stickyref, $setstickyref ) {
 		    if ( $chainref->{name} eq 'sticky' ) {
 			$rule1 = $_;
-			$rule1 =~ s/-j sticky/-m recent --name $list --update --seconds 300 -j MARK --set-mark $mark/;
+
+			set_rule_target( $rule1, 'MARK',   "--set-mark $mark" );
+			set_rule_option( $rule1, 'recent', "--name $list --update --seconds 300" );
+
 			$rule2 = $_;
-			$rule2 =~ s/-j sticky/-m mark --mark 0\/$mask -m recent --name $list --remove/;
+
+			clear_rule_target( $rule2 );
+			set_rule_option( $rule2, 'mark', "--mark 0/$mask -m recent --name $list --remove" );
 		    } else {
 			$rule1 = $_;
-			$rule1 =~ s/-j sticky/-m mark --mark $mark\/$mask -m recent --name $list --set/;
+
+			clear_rule_target( $rule1 );
+			set_rule_option( $rule1, 'mark', "--mark $mark\/$mask -m recent --name $list --set" ); 
+
 			$rule2 = '';
 		    }
 
-		    assert ( $rule1 =~ s/^-A // );
-		    add_rule $chainref, $rule1;
+		    add_trule $chainref, $rule1;
 
 		    if ( $rule2 ) {
-			assert ( $rule2 =~ s/^-A // );
-			add_rule $chainref, $rule2;
+			add_trule $chainref, $rule2;
 		    }
 		}
 	    }
 
-	    for ( grep /-j sticko/, @{$tcoutref->{rules}} ) {
+	    for ( grep rule_target( $_ ) eq 'sticko', , @{$tcoutref->{rules}} ) {
 		my ( $rule1, $rule2 );
 		my $list = sprintf "sticky%03d" , $sticky++;
 		my $stickoref = ensure_mangle_chain 'sticko';
@@ -1103,21 +1346,27 @@ sub handle_stickiness( $ ) {
 		for my $chainref ( $stickoref, $setstickoref ) {
 		    if ( $chainref->{name} eq 'sticko' ) {
 			$rule1 = $_;
-			$rule1 =~ s/-j sticko/-m recent --name $list --rdest --update --seconds 300 -j MARK --set-mark $mark/;
+
+			set_rule_target( $rule1, 'MARK',   "--set-mark $mark" );
+			set_rule_option( $rule1, 'recent', " --name $list --rdest --update --seconds 300 -j MARK --set-mark $mark" );
+
 			$rule2 = $_;
-			$rule2 =~ s/-j sticko/-m mark --mark 0\/$mask -m recent --name $list --rdest --remove/;
+			
+			clear_rule_target( $rule2 );
+			set_rule_option  ( $rule2, 'mark', "--mark 0\/$mask -m recent --name $list --rdest --remove" );
 		    } else {
 			$rule1 = $_;
-			$rule1 =~ s/-j sticko/-m mark --mark $mark -m recent --name $list --rdest --set/;
+
+			clear_rule_target( $rule1 );
+			set_rule_option  ( $rule1, 'mark', "--mark $mark -m recent --name $list --rdest --set" );
+
 			$rule2 = '';
 		    }
 
-		    assert( $rule1 =~ s/-A // );
-		    add_rule $chainref, $rule1;
+		    add_trule $chainref, $rule1;
 
 		    if ( $rule2 ) {
-			$rule2 =~ s/-A //;
-			add_rule $chainref, $rule2;
+			add_trule $chainref, $rule2;
 		    }
 		}
 	    }
